@@ -33,8 +33,7 @@ class AppleFrameworks(ExternalDependency):
         for f in self.frameworks:
             self.link_args += ['-framework', f]
 
-    def found(self):
-        return mesonlib.is_osx()
+        self.is_found = mesonlib.for_darwin(self.want_cross, self.env)
 
-    def get_version(self):
-        return 'unknown'
+    def log_tried(self):
+        return 'framework'
