@@ -14,11 +14,15 @@
 
 """A library of random helper functionality."""
 
+import functools
 import sys
 import stat
 import time
 import platform, subprocess, operator, os, shutil, re
 import collections
+from enum import Enum
+from functools import lru_cache
+
 from mesonbuild import mlog
 
 have_fcntl = False
@@ -47,6 +51,23 @@ else:
     python_command = [sys.executable]
 meson_command = None
 
+def set_meson_command(mainfile):
+    global python_command
+    global meson_command
+    # On UNIX-like systems `meson` is a Python script
+    # On Windows `meson` and `meson.exe` are wrapper exes
+    if not mainfile.endswith('.py'):
+        meson_command = [mainfile]
+    elif os.path.isabs(mainfile) and mainfile.endswith('mesonmain.py'):
+        # Can't actually run meson with an absolute path to mesonmain.py, it must be run as -m mesonbuild.mesonmain
+        meson_command = python_command + ['-m', 'mesonbuild.mesonmain']
+    else:
+        # Either run uninstalled, or full path to meson-script.py
+        meson_command = python_command + [mainfile]
+    # We print this value for unit tests.
+    if 'MESON_COMMAND_TESTS' in os.environ:
+        mlog.log('meson_command is {!r}'.format(meson_command))
+
 def is_ascii_string(astring):
     try:
         if isinstance(astring, str):
@@ -83,6 +104,13 @@ an_unpicklable_object = threading.Lock()
 class MesonException(Exception):
     '''Exceptions thrown by Meson'''
 
+    def get_msg_with_context(self):
+        s = ''
+        if hasattr(self, 'lineno') and hasattr(self, 'file'):
+            s = get_error_location_string(self.file, self.lineno) + ' '
+        s += str(self)
+        return s
+
 class EnvironmentException(MesonException):
     '''Exceptions thrown while processing and creating the build environment'''
 
@@ -196,6 +224,7 @@ class File:
         return ret.format(self.relative_name())
 
     @staticmethod
+    @lru_cache(maxsize=None)
     def from_source_file(source_root, subdir, fname):
         if not os.path.isfile(os.path.join(source_root, subdir, fname)):
             raise MesonException('File %s does not exist.' % fname)
@@ -209,12 +238,14 @@ class File:
     def from_absolute_file(fname):
         return File(False, '', fname)
 
+    @lru_cache(maxsize=None)
     def rel_to_builddir(self, build_to_src):
         if self.is_built:
             return self.relative_name()
         else:
             return os.path.join(build_to_src, self.subdir, self.fname)
 
+    @lru_cache(maxsize=None)
     def absolute_path(self, srcdir, builddir):
         absdir = srcdir
         if self.is_built:
@@ -233,6 +264,7 @@ class File:
     def __hash__(self):
         return hash((self.fname, self.subdir, self.is_built))
 
+    @lru_cache(maxsize=None)
     def relative_name(self):
         return os.path.join(self.subdir, self.fname)
 
@@ -240,7 +272,7 @@ def get_compiler_for_source(compilers, src):
     for comp in compilers:
         if comp.can_compile(src):
             return comp
-    raise RuntimeError('No specified compiler can handle file {!s}'.format(src))
+    raise MesonException('No specified compiler can handle file {!s}'.format(src))
 
 def classify_unity_sources(compilers, sources):
     compsrclist = {}
@@ -252,6 +284,53 @@ def classify_unity_sources(compilers, sources):
             compsrclist[comp].append(src)
     return compsrclist
 
+class OrderedEnum(Enum):
+    """
+    An Enum which additionally offers homogeneous ordered comparison.
+    """
+    def __ge__(self, other):
+        if self.__class__ is other.__class__:
+            return self.value >= other.value
+        return NotImplemented
+
+    def __gt__(self, other):
+        if self.__class__ is other.__class__:
+            return self.value > other.value
+        return NotImplemented
+
+    def __le__(self, other):
+        if self.__class__ is other.__class__:
+            return self.value <= other.value
+        return NotImplemented
+
+    def __lt__(self, other):
+        if self.__class__ is other.__class__:
+            return self.value < other.value
+        return NotImplemented
+
+MachineChoice = OrderedEnum('MachineChoice', ['BUILD', 'HOST', 'TARGET'])
+
+class PerMachine:
+    def __init__(self, build, host, target):
+        self.build = build
+        self.host = host
+        self.target = target
+
+    def __getitem__(self, machine: MachineChoice):
+        return {
+            MachineChoice.BUILD:  self.build,
+            MachineChoice.HOST:   self.host,
+            MachineChoice.TARGET: self.target
+        }[machine]
+
+    def __setitem__(self, machine: MachineChoice, val):
+        key = {
+            MachineChoice.BUILD:  'build',
+            MachineChoice.HOST:   'host',
+            MachineChoice.TARGET: 'target'
+        }[machine]
+        setattr(self, key, val)
+
 def is_osx():
     return platform.system().lower() == 'darwin'
 
@@ -284,77 +363,93 @@ def is_dragonflybsd():
 def is_freebsd():
     return platform.system().lower() == 'freebsd'
 
+def _get_machine_is_cross(env, is_cross):
+    """
+    This is not morally correct, but works for now. For cross builds the build
+    and host machines differ. `is_cross == true` means the host machine, while
+    `is_cross == false` means the build machine. Both are used in practice,
+    even though the documentation refers to the host machine implying we should
+    hard-code it. For non-cross builds `is_cross == false` is passed but the
+    host and build machines are identical so it doesn't matter.
+
+    Users for `for_*` should instead specify up front which machine they want
+    and query that like:
+
+        env.machines[MachineChoice.HOST].is_haiku()
+
+    """
+    for_machine = MachineChoice.HOST if is_cross else MachineChoice.BUILD
+    return env.machines[for_machine]
+
 def for_windows(is_cross, env):
     """
     Host machine is windows?
 
+    Deprecated: Please use `env.machines[for_machine].is_windows()`.
+
     Note: 'host' is the machine on which compiled binaries will run
     """
-    if not is_cross:
-        return is_windows()
-    elif env.cross_info.has_host():
-        return env.cross_info.config['host_machine']['system'] == 'windows'
-    return False
+    return _get_machine_is_cross(env, is_cross).is_windows()
 
 def for_cygwin(is_cross, env):
     """
     Host machine is cygwin?
 
+    Deprecated: Please use `env.machines[for_machine].is_cygwin()`.
+
     Note: 'host' is the machine on which compiled binaries will run
     """
-    if not is_cross:
-        return is_cygwin()
-    elif env.cross_info.has_host():
-        return env.cross_info.config['host_machine']['system'] == 'cygwin'
-    return False
+    return _get_machine_is_cross(env, is_cross).is_cygwin()
 
 def for_linux(is_cross, env):
     """
     Host machine is linux?
 
+    Deprecated: Please use `env.machines[for_machine].is_linux()`.
+
     Note: 'host' is the machine on which compiled binaries will run
     """
-    if not is_cross:
-        return is_linux()
-    elif env.cross_info.has_host():
-        return env.cross_info.config['host_machine']['system'] == 'linux'
-    return False
+    return _get_machine_is_cross(env, is_cross).is_linux()
 
 def for_darwin(is_cross, env):
     """
     Host machine is Darwin (iOS/OS X)?
 
+    Deprecated: Please use `env.machines[for_machine].is_darwin()`.
+
     Note: 'host' is the machine on which compiled binaries will run
     """
-    if not is_cross:
-        return is_osx()
-    elif env.cross_info.has_host():
-        return env.cross_info.config['host_machine']['system'] in ('darwin', 'ios')
-    return False
+    return _get_machine_is_cross(env, is_cross).is_darwin()
 
 def for_android(is_cross, env):
     """
     Host machine is Android?
 
+    Deprecated: Please use `env.machines[for_machine].is_android()`.
+
     Note: 'host' is the machine on which compiled binaries will run
     """
-    if not is_cross:
-        return is_android()
-    elif env.cross_info.has_host():
-        return env.cross_info.config['host_machine']['system'] == 'android'
-    return False
+    return _get_machine_is_cross(env, is_cross).is_android()
 
 def for_haiku(is_cross, env):
     """
     Host machine is Haiku?
 
+    Deprecated: Please use `env.machines[for_machine].is_haiku()`.
+
     Note: 'host' is the machine on which compiled binaries will run
     """
-    if not is_cross:
-        return is_haiku()
-    elif env.cross_info.has_host():
-        return env.cross_info.config['host_machine']['system'] == 'haiku'
-    return False
+    return _get_machine_is_cross(env, is_cross).is_haiku()
+
+def for_openbsd(is_cross, env):
+    """
+    Host machine is OpenBSD?
+
+    Deprecated: Please use `env.machines[for_machine].is_openbsd()`.
+
+    Note: 'host' is the machine on which compiled binaries will run
+    """
+    return _get_machine_is_cross(env, is_cross).is_openbsd()
 
 def exe_exists(arglist):
     try:
@@ -383,33 +478,59 @@ def detect_vcs(source_dir):
                 return vcs
     return None
 
-def grab_leading_numbers(vstr, strict=False):
-    result = []
-    for x in vstr.rstrip('.').split('.'):
-        try:
-            result.append(int(x))
-        except ValueError as e:
-            if strict:
-                msg = 'Invalid version to compare against: {!r}; only ' \
-                      'numeric digits separated by "." are allowed: ' + str(e)
-                raise MesonException(msg.format(vstr))
-            break
-    return result
+# a helper class which implements the same version ordering as RPM
+@functools.total_ordering
+class Version:
+    def __init__(self, s):
+        self._s = s
 
-def make_same_len(listA, listB):
-    maxlen = max(len(listA), len(listB))
-    for i in listA, listB:
-        for n in range(len(i), maxlen):
-            i.append(0)
+        # split into numeric, alphabetic and non-alphanumeric sequences
+        sequences = re.finditer(r'(\d+|[a-zA-Z]+|[^a-zA-Z\d]+)', s)
+        # non-alphanumeric separators are discarded
+        sequences = [m for m in sequences if not re.match(r'[^a-zA-Z\d]+', m.group(1))]
+        # numeric sequences have leading zeroes discarded
+        sequences = [re.sub(r'^0+(\d)', r'\1', m.group(1), 1) for m in sequences]
 
-numpart = re.compile('[0-9.]+')
+        self._v = sequences
 
-def version_compare(vstr1, vstr2, strict=False):
-    match = numpart.match(vstr1.strip())
-    if match is None:
-        msg = 'Uncomparable version string {!r}.'
-        raise MesonException(msg.format(vstr1))
-    vstr1 = match.group(0)
+    def __str__(self):
+        return '%s (V=%s)' % (self._s, str(self._v))
+
+    def __lt__(self, other):
+        return self.__cmp__(other) == -1
+
+    def __eq__(self, other):
+        return self.__cmp__(other) == 0
+
+    def __cmp__(self, other):
+        def cmp(a, b):
+            return (a > b) - (a < b)
+
+        # compare each sequence in order
+        for i in range(0, min(len(self._v), len(other._v))):
+            # sort a non-digit sequence before a digit sequence
+            if self._v[i].isdigit() != other._v[i].isdigit():
+                return 1 if self._v[i].isdigit() else -1
+
+            # compare as numbers
+            if self._v[i].isdigit():
+                # because leading zeros have already been removed, if one number
+                # has more digits, it is greater
+                c = cmp(len(self._v[i]), len(other._v[i]))
+                if c != 0:
+                    return c
+                # fallthrough
+
+            # compare lexicographically
+            c = cmp(self._v[i], other._v[i])
+            if c != 0:
+                return c
+
+        # if equal length, all components have matched, so equal
+        # otherwise, the version with a suffix remaining is greater
+        return cmp(len(self._v), len(other._v))
+
+def _version_extract_cmpop(vstr2):
     if vstr2.startswith('>='):
         cmpop = operator.ge
         vstr2 = vstr2[2:]
@@ -433,10 +554,12 @@ def version_compare(vstr1, vstr2, strict=False):
         vstr2 = vstr2[1:]
     else:
         cmpop = operator.eq
-    varr1 = grab_leading_numbers(vstr1, strict)
-    varr2 = grab_leading_numbers(vstr2, strict)
-    make_same_len(varr1, varr2)
-    return cmpop(varr1, varr2)
+
+    return (cmpop, vstr2)
+
+def version_compare(vstr1, vstr2):
+    (cmpop, vstr2) = _version_extract_cmpop(vstr2)
+    return cmpop(Version(vstr1), Version(vstr2))
 
 def version_compare_many(vstr1, conditions):
     if not isinstance(conditions, (list, tuple, frozenset)):
@@ -444,28 +567,22 @@ def version_compare_many(vstr1, conditions):
     found = []
     not_found = []
     for req in conditions:
-        if not version_compare(vstr1, req, strict=True):
+        if not version_compare(vstr1, req):
             not_found.append(req)
         else:
             found.append(req)
     return not_found == [], not_found, found
 
-
+# determine if the minimum version satisfying the condition |condition| exceeds
+# the minimum version for a feature |minimum|
 def version_compare_condition_with_min(condition, minimum):
-    match = numpart.match(minimum.strip())
-    if match is None:
-        msg = 'Uncomparable version string {!r}.'
-        raise MesonException(msg.format(minimum))
-    minimum = match.group(0)
     if condition.startswith('>='):
         cmpop = operator.le
         condition = condition[2:]
     elif condition.startswith('<='):
-        return True
-        condition = condition[2:]
+        return False
     elif condition.startswith('!='):
-        return True
-        condition = condition[2:]
+        return False
     elif condition.startswith('=='):
         cmpop = operator.le
         condition = condition[2:]
@@ -476,49 +593,25 @@ def version_compare_condition_with_min(condition, minimum):
         cmpop = operator.lt
         condition = condition[1:]
     elif condition.startswith('<'):
-        return True
-        condition = condition[2:]
+        return False
     else:
         cmpop = operator.le
-    varr1 = grab_leading_numbers(minimum, True)
-    varr2 = grab_leading_numbers(condition, True)
-    make_same_len(varr1, varr2)
-    return cmpop(varr1, varr2)
 
-def version_compare_condition_with_max(condition, maximum):
-    match = numpart.match(maximum.strip())
-    if match is None:
-        msg = 'Uncomparable version string {!r}.'
-        raise MesonException(msg.format(maximum))
-    maximum = match.group(0)
-    if condition.startswith('>='):
-        return False
-        condition = condition[2:]
-    elif condition.startswith('<='):
-        cmpop = operator.ge
-        condition = condition[2:]
-    elif condition.startswith('!='):
-        return False
-        condition = condition[2:]
-    elif condition.startswith('=='):
-        cmpop = operator.ge
-        condition = condition[2:]
-    elif condition.startswith('='):
-        cmpop = operator.ge
-        condition = condition[1:]
-    elif condition.startswith('>'):
-        return False
-        condition = condition[1:]
-    elif condition.startswith('<'):
-        cmpop = operator.gt
-        condition = condition[2:]
-    else:
-        cmpop = operator.ge
-    varr1 = grab_leading_numbers(maximum, True)
-    varr2 = grab_leading_numbers(condition, True)
-    make_same_len(varr1, varr2)
-    return cmpop(varr1, varr2)
+    # Declaring a project(meson_version: '>=0.46') and then using features in
+    # 0.46.0 is valid, because (knowing the meson versioning scheme) '0.46.0' is
+    # the lowest version which satisfies the constraint '>=0.46'.
+    #
+    # But this will fail here, because the minimum version required by the
+    # version constraint ('0.46') is strictly less (in our version comparison)
+    # than the minimum version needed for the feature ('0.46.0').
+    #
+    # Map versions in the constraint of the form '0.46' to '0.46.0', to embed
+    # this knowledge of the meson versioning scheme.
+    condition = condition.strip()
+    if re.match('^\d+.\d+$', condition):
+        condition += '.0'
 
+    return cmpop(Version(minimum), Version(condition))
 
 def default_libdir():
     if is_debianlike():
@@ -603,7 +696,9 @@ def do_replacement(regex, line, format, confdata):
                 elif isinstance(var, int):
                     var = str(var)
                 else:
-                    raise RuntimeError('Tried to replace a variable with something other than a string or int.')
+                    msg = 'Tried to replace variable {!r} value with ' \
+                          'something other than a string or int: {!r}'
+                    raise MesonException(msg.format(varname, var))
             else:
                 missing_variables.add(varname)
                 var = ''
@@ -653,12 +748,18 @@ def do_conf_file(src, dst, confdata, format, encoding='utf-8'):
 
     result = []
     missing_variables = set()
+    # Detect when the configuration data is empty and no tokens were found
+    # during substitution so we can warn the user to use the `copy:` kwarg.
+    confdata_useless = not confdata.keys()
     for line in data:
         if line.startswith(search_token):
+            confdata_useless = False
             line = do_mesondefine(line, confdata)
         else:
             line, missing = do_replacement(regex, line, format, confdata)
             missing_variables.update(missing)
+            if missing:
+                confdata_useless = False
         result.append(line)
     dst_tmp = dst + '~'
     try:
@@ -668,7 +769,7 @@ def do_conf_file(src, dst, confdata, format, encoding='utf-8'):
         raise MesonException('Could not write output file %s: %s' % (dst, str(e)))
     shutil.copymode(src, dst_tmp)
     replace_if_different(dst, dst_tmp)
-    return missing_variables
+    return missing_variables, confdata_useless
 
 CONF_C_PRELUDE = '''/*
  * Autogenerated by the Meson build system.
@@ -1015,6 +1116,22 @@ def windows_proof_rmtree(f):
     shutil.rmtree(f)
 
 
+def windows_proof_rm(fpath):
+    """Like windows_proof_rmtree, but for a single file."""
+    if os.path.isfile(fpath):
+        os.chmod(fpath, os.stat(fpath).st_mode | stat.S_IWRITE | stat.S_IREAD)
+    delays = [0.1, 0.1, 0.2, 0.2, 0.2, 0.5, 0.5, 1, 1, 1, 1, 2]
+    for d in delays:
+        try:
+            os.unlink(fpath)
+            return
+        except FileNotFoundError:
+            return
+        except (OSError, PermissionError):
+            time.sleep(d)
+    os.unlink(fpath)
+
+
 def detect_subprojects(spdir_name, current_dir='', result=None):
     if result is None:
         result = {}
@@ -1039,7 +1156,16 @@ def detect_subprojects(spdir_name, current_dir='', result=None):
                 result[basename] = [trial]
     return result
 
-class OrderedSet(collections.MutableSet):
+def get_error_location_string(fname, lineno):
+    return '{}:{}:'.format(fname, lineno)
+
+def substring_is_in_list(substr, strlist):
+    for s in strlist:
+        if substr in s:
+            return True
+    return False
+
+class OrderedSet(collections.abc.MutableSet):
     """A set that preserves the order in which items are added, by first
     insertion.
     """
@@ -1064,6 +1190,9 @@ class OrderedSet(collections.MutableSet):
                 '", "'.join(repr(e) for e in self.__container.keys()))
         return 'OrderedSet()'
 
+    def __reversed__(self):
+        return reversed(self.__container)
+
     def add(self, value):
         self.__container[value] = None
 
@@ -1100,3 +1229,12 @@ class BuildDirLock:
         elif have_msvcrt:
             msvcrt.locking(self.lockfile.fileno(), msvcrt.LK_UNLCK, 1)
         self.lockfile.close()
+
+def relpath(path, start):
+    # On Windows a relative path can't be evaluated for paths on two different
+    # drives (i.e. c:\foo and f:\bar).  The only thing left to do is to use the
+    # original absolute path.
+    try:
+        return os.path.relpath(path, start)
+    except ValueError:
+        return path
